/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*-
 */
#include <hash_map>
/*
  #ifndef lint
  static const char rcsid[] =
  "@(#) $Header: /nfs/jade/vint/CVSROOT/ns-2/queue/drr.cc,v 1.9 2000/09/01 03:04:05 haoboy Exp $ (Xerox)";
  #endif
*/

//#include "ns-indlude/config.h"   // for string.h
#include "config.h"   // for string.h
#include <stdlib.h>
//#include "ns-include/queue/queue.h"
#include "queue/queue.h"

class BinBSFQ;
class BSFQ;
 
struct flow {
	double prevTime;
	int src;
	int weight;
};

typedef struct flow* flow_s;

class BinBSFQ : public PacketQueue {
	BinBSFQ(): src(-1),prev(0),next(0),PacketQueue() {}
	friend BSFQ;
	protected :
	int pkts;
	int src;    //to detect collisions keep track of actual src address
	//t bcount; //count of bytes in each flow to find the max flow;
	BinBSFQ *prev;
	BinBSFQ *next;
	//int bin_num;// to which delta this belongs...
	/*
	  inline BinBSFQ * activate(BinBSFQ *head) {
	  if (head) {
	  this->prev = head->prev;
	  this->next = head;
	  head->prev->next = this;
	  head->prev = this;
	  return head;
	  }
	  this->prev = this;
	  this->next = this;
	  return this;
	  }
	  inline BinBSFQ * idle(BinBSFQ *head) {
	  if (head == this) {
	  if (this->next == this)
	  return 0;
	  this->next->prev = this->prev;
	  this->prev->next = this->next;
	  return this->next;
	  }
	  this->next->prev = this->prev;
	  this->prev->next = this->next;
	  return head;
	  }
	*/
};

class BSFQ : public Queue {
	public :
	BSFQ();
	virtual int command(int argc, const char*const* argv);
	Packet *deque(void);
	void enque(Packet *pkt);
	int hash(Packet *pkt);
	int hash(int i); 
	void clear();
protected:
	int num_bins_ ; //total number of flows allowed
	int mask_; // for src
	int cur;
	double curTime;
	int delta_;
	//	int blimit_;    //total number of bytes allowed across all flows
	int bytecnt ; //cumulative sum of bytes across all flows
	int pktcnt ; // cumulative sum of packets across all flows
	int flwcnt_ ; //total number of active flows
	BinBSFQ* bins; //pointer to the entire drr struct
	hash_map<int, flow_s> flowTable;
  
public:
	
	//returns queuelength in packets
	inline int length () {
		return pktcnt;
	}
	  
	//returns queuelength in bytes
 	inline int blength () {
		return bytecnt;
	}
	
};

static class BSFQClass : public TclClass {
public:
	BSFQClass() : TclClass("Queue/BSFQ") {}
	TclObject* create(int, const char*const*) {
		return (new BSFQ());
	}
} class_bsfq;

BSFQ::BSFQ() : Queue()
{
	num_bins_=16;
	cur=0;
	curTime = 0;
	flwcnt_=0;
	bytecnt=0;
	pktcnt=0;
	mask_=0;
	bins = new BinBSFQ[num_bins_];
	cout << "init flowTable\n";
	flowTable = *(new hash_map<int, flow_s>::hash_map());
	bind("num_bins_",&num_bins_);
	//	bind("blimit_",&blimit_);
	bind("mask_",&mask_);
	bind("delta__",&delta_);
	bind("flwcnt_",&flwcnt_);
}

void BSFQ::enque(Packet* pkt)
{
	cout << "entering enque\n";
	flow_s fl;
	int fhash= hash(pkt) % flwcnt_;
	fl = flowTable[fhash];

	hdr_ip *iph=hdr_ip::access(pkt);
	hdr_cmn *ch= hdr_cmn::access(pkt);	
	/*detect collisions here */
	int compare=(!mask_ ? ((int)iph->saddr()) : ((int)iph->saddr()&0xfff0));
	if (fl->src ==-1)
		fl->src=compare;
	else
		if (fl->src != compare)
			fprintf(stderr,"Collisions between %d and %d src addresses\n",fl->src,(int)iph->saddr());      

	double ts = ch->timestamp();
	double vts = max(curTime,fl->prevTime) + ch->size()/(fl->weight+0.0);
	int which = (int)(vts - curTime)/delta_;
	cout << "curTime="<<curTime<<" prevTime = "<<fl->prevTime<<" which="<<which<<" vts="<<vts<<endl;
	fl->prevTime = vts;

	if (which < num_bins_) {
		BinBSFQ* q = &bins[(cur+which) % num_bins_];
		cout << "which = "<<which<<" num_bins_="<<num_bins_<< "q->pkts=" << q->pkts << endl;
		q->enque(pkt);
		++(q->pkts);
		++pktcnt;
		bytecnt += ch->size();
		cout << "++++which = "<<which<<" num_bins_="<<num_bins_<< "q->pkts=" << q->pkts << endl;
	} else {
		//drop
		drop(pkt);
	}
	cout << "leaving enque\n";
}

Packet *BSFQ::deque(void) 
{
	cout << "entering deque\n";
	BinBSFQ* bin = &bins[cur];
	Packet* pkt;
	//	if (bytecnt==0||pktcnt==0) 
	//return(0);

	for (int i=0;i<num_bins_;++i) {
		cout <<" cur=="<<cur<<" bin->pkts==" << bin->pkts << endl;
		if (bin->pkts==0) {//if this bin is
			cur = (cur+1) % num_bins_;
			curTime += delta_;
			bin = &bins[cur];
		} else {
			cout <<"debug\n";
			pkt = bin->deque();
			--pktcnt;
			--bin->pkts;
			hdr_cmn *ch= hdr_cmn::access(pkt);	
			bytecnt -= ch->size();
			return pkt;
		}
	}
	return 0;//not reached
	cout << "leaving deque\n";
}

void BSFQ::clear()
{
	BinBSFQ bin ;
	for (int i=0;i<num_bins_;++i) {
		if (bins[i].length()!=0) {
			fprintf(stderr, "Changing non-empty bucket from bins\n");
			exit(1);
		}
	}
	delete[](bins);
	bins = 0;
}

/*
 *Allows one to change blimit_ and bucket_ for a particular binsQ :
 *
 *
 */
int BSFQ::command(int argc, const char*const* argv)
{
	if (argc==3) {
		if (strcmp(argv[1], "delta") == 0) {
			delta_ = atoi(argv[2]);
			return (TCL_OK);
		}
		if (strcmp(argv[1], "flwcnt") == 0) {
			flwcnt_ = atoi(argv[2]);
			return (TCL_OK);
		}
		if (strcmp(argv[1], "num_bins") == 0) {
			clear();
			num_bins_ = atoi(argv[2]);
			cout << "init binsQ\n";
			bins = new BinBSFQ[num_bins_];
			return (TCL_OK);
		}
		if (strcmp(argv[1],"mask")==0) {
			mask_= atoi(argv[2]);
			return (TCL_OK);
		}
	}
	if (argc==4) {
		if (strcmp(argv[1],"quantum") == 0) {
			int index__ = atoi(argv[2]);
			cout << "init weight\n";
			flowTable[index__] = new flow;
			flowTable[index__]->weight = atoi(argv[3]);
			flowTable[index__]->prevTime = 0;
			cout << "done weight\n";
			return (TCL_OK);
		}
	}
	return (Queue::command(argc, argv));
}

int BSFQ::hash(Packet* pkt)
{
	hdr_ip *iph=hdr_ip::access(pkt);
	int i;
	if (mask_)
		i = (int)iph->saddr() & (0xfff0);
	else
		i = (int)iph->saddr();
	return ((i + (i >> 7) + ~(i>>4)) % ((2<<23)-1))+1; // modulo a large prime
}

int BSFQ::hash(int i) 
{
	return ((i + (i >> 8) + ~(i>>4)) % ((2<<23)-1))+1; // modulo a large prime
}



