/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*-
 */
/*
#ifndef lint
static const char rcsid[] =
"@(#) $Header: /nfs/jade/vint/CVSROOT/ns-2/queue/wfq.cc,v 1.9 2000/09/01 03:04:05 haoboy Exp $ (Xerox)";
#endif
*/
#include "config.h"   // for string.h
#include <stdlib.h>
#include <values.h>
#include <iostream.h>
#include "queue.h"

class PacketWFQ;
class WFQ;

class PacketWFQ : public PacketQueue {
	PacketWFQ(): pkts(0),src(-1),bcount(0),prev(0),next(0),head_finish_time(0) {}
	friend WFQ;
public:
	double weight; //Weight assigned to this queue. bandwidth alotted = weight*totalbandwidth
	double head_finish_time;//Finish time of the packet at head of this queue.
protected :
	int pkts;
	int src;    //to detect collisions keep track of actual src address
	int bcount; //count of bytes in each flow to find the max flow;
	PacketWFQ *prev;
	PacketWFQ *next;
};

class WFQ : public Queue {
public :
	WFQ();
	virtual int command(int argc, const char*const* argv);
	Packet *deque(void);
	void enque(Packet *pkt);
	int hash(Packet *pkt);
	void clear();
protected:
	int buckets_ ; //total number of flows allowed
	int blimit_;    //total number of bytes allowed across all flows
	int mask_;     /*if set hashes on just the node address otherwise on 
			 node+port address*/
	int bytecnt ; //cumulative sum of bytes across all flows
	int pktcnt ; // cumulative sum of packets across all flows
	int flwcnt ; //total number of active flows
	double curr_time; //Finish time assigned to last packet
	double bandwidth_; //Total Bandwidth of the link in bytespersecond
	PacketWFQ *curr; //current active flow
	PacketWFQ *wfq ; //pointer to the entire wfq struct
	
	
	inline PacketWFQ *getMaxflow (PacketWFQ *curr) { //returns flow with max pkts
		int bcount = 0;
		int maxflow=0;
		for(int i=0;i<buckets_;i++) {
			if (wfq[maxflow].bcount < wfq[i].bcount)
				maxflow=i;
		}
		return &(wfq[maxflow]);
	}
	
public:
	//returns queuelength in packets
	inline int length () {
		return pktcnt;
	}
	
	//returns queuelength in bytes
	inline int blength () {
		return bytecnt;
	}
};

static class WFQClass : public TclClass {
public:
	WFQClass() : TclClass("Queue/WFQ") {}
	TclObject* create(int, const char*const*) {
		return (new WFQ());
	}
} class_wfq;

WFQ::WFQ()
{
	buckets_=3;
	wfq=0;
	blimit_ = 1000;
	curr=0;
	flwcnt=0;
	bytecnt=0;
	pktcnt=0;
	curr_time=0;
	mask_=0;
	bandwidth_=1000000;
	bind("buckets_",&buckets_);
	bind("blimit_",&blimit_);
	bind("mask_",&mask_);
	bind("bandwidth_",&bandwidth_);
}

void WFQ::enque(Packet* pkt)
{
	//cout<< "HERE 3 buckets_ = "<< buckets_<<"\n";
	PacketWFQ *q,*remq;
	int which;
	
	hdr_cmn *ch= hdr_cmn::access(pkt);
	hdr_ip *iph = hdr_ip::access(pkt);
	if (!wfq)
		wfq=new PacketWFQ[buckets_];
	which= hash(pkt) % buckets_;
	q=&wfq[which];
	//cout<< "HERE 3.5\n";
	/*detect collisions here */
	
	int compare=(!mask_ ? ((int)iph->saddr()) : ((int)iph->saddr()&0xfff0));
	if (q->src ==-1)
		q->src=compare;
	else
		if (q->src != compare)
			fprintf(stderr,"Collisions between %d and %d src addresses\n",q->src,(int)iph->saddr());      
	
	//cout<< "HERE 3.75\n";
	q->enque(pkt);
	++q->pkts;
	++pktcnt;
	q->bcount += ch->size();
	bytecnt +=ch->size();
	//cout<< "HERE 4\n";

	if (q->pkts==1) {
		++flwcnt;
		/* As this is the first packet in this queue assign the finish_time */
		//cout<<"q->weight = "<<q->weight<<", bandwidth_ = " << bandwidth_ <<"\n"; 
		//q->head_finish_time = curr_time + (double)ch->size()/(double)(q->weight * bandwidth_) ;
		//curr_time = q->head_finish_time;
		q->head_finish_time = curr_time + (double)ch->size()/(double)(q->weight * bandwidth_) ;
		curr_time = q->head_finish_time;
	}
	//cout<< "HERE 5\n";
	/*Drop packets if the queue is at above capacity*/
	while (bytecnt > blimit_) {
		Packet *p;
		//cout<< "HERE 5.1\n";
		hdr_cmn *remch;
		hdr_ip *remiph;
		remq=getMaxflow(curr);
		//cout<< "HERE 5.2\n";
		if(remq) {
			//cout<< "HERE 5.3\n";
			p=remq->deque();
			remch=hdr_cmn::access(p);
			remiph=hdr_ip::access(p);
			//cout<< "HERE 5.4\n";
			remq->bcount -= remch->size();
			bytecnt -= remch->size();
			drop(p);
			//cout<< "HERE 5.5\n";
			--remq->pkts;
			--pktcnt;
			if (remq->pkts==0) {
				//curr=remq->idle(curr);
				--flwcnt;
			}
		}
		
	}
	//cout<< "HERE 6\n";
}

Packet *WFQ::deque(void) {
	int minQ = 0;
	if (bytecnt==0) {
		fprintf (stderr,"No active flow\n");
		return(0);
	}
	/*Find Q with the minimum head_finish_time*/
	for(int i=0;i<buckets_;i++) {
		if (wfq[i].pkts > 1) {
			if(wfq[i].head_finish_time < wfq[minQ].head_finish_time) minQ = i;
			//curr_time = wfq[i].head_finish_time;
		}
	}
	
	if(wfq[minQ].pkts<1) {
		fprintf (stderr,"No active flow\n");
		return(0);
	}
	hdr_cmn *ch;
	hdr_ip *iph;
	Packet *pkt=0;
	pkt=wfq[minQ].lookup(0);  
	ch=hdr_cmn::access(pkt);
	iph=hdr_ip::access(pkt);
	pkt = wfq[minQ].deque();
	wfq[minQ].bcount -= ch->size();
	--wfq[minQ].pkts;
	--pktcnt;
	bytecnt -= ch->size();
	
	//cout<< "HERE 1: minQ = "<< minQ<<"\n ";
	if(wfq[minQ].pkts==0) {
		--flwcnt;
		wfq[minQ].head_finish_time = MAXDOUBLE;
		//wfq[minQ] = *(wfq[minQ].idle(&(wfq[minQ])));
	} else {
		/*Re calculate the head_finish_time for this queue*/
		Packet *pkt2 = wfq[minQ].lookup(0);
		ch=hdr_cmn::access(pkt2);
		iph=hdr_ip::access(pkt2);
		//cout<<"q->weight = "<<wfq[minQ].weight <<", bandwidth_ = " << bandwidth_ <<"\n"; 
		//wfq[minQ].head_finish_time = curr_time + (double)ch->size()/(double)(wfq[minQ].weight * bandwidth_) ;
		//curr_time = wfq[minQ].head_finish_time;
		wfq[minQ].head_finish_time = wfq[minQ].head_finish_time + (double)ch->size()/(double)(wfq[minQ].weight * bandwidth_) ;
		//cout << "curr_time = " << curr_time <<"\n";
		if(curr_time < wfq[minQ].head_finish_time) curr_time = wfq[minQ].head_finish_time;
		//cout <<"curr_time = " << wfq[minQ].head_finish_time <<"\n";
	}
	
	//cout<< "HERE 2\n";
	return pkt;
}

void WFQ::clear()
{
	PacketWFQ *q =wfq;
	int i = buckets_;

	if (!q)
		return;
	while (i--) {
		if (q->pkts) {
			fprintf(stderr, "Changing non-empty bucket from wfq\n");
			exit(1);
		}
		++q;
	}
	delete[](wfq);
	wfq = 0;
}

/*
 *Allows one to change blimit_ and bucket_ for a particular wfqQ :
 *
 *
 */
int WFQ::command(int argc, const char*const* argv)
{
	if (argc==3) {
		if (strcmp(argv[1], "blimit") == 0) {
			blimit_ = atoi(argv[2]);
			if (bytecnt > blimit_)
				{
					fprintf (stderr,"More packets in buffer than the new limit");
					exit (1);
				}
			return (TCL_OK);
		}
		if (strcmp(argv[1], "buckets") == 0) {
			clear();
			buckets_ = atoi(argv[2]);
			//cout <<"Setting buckets_ = " << buckets_;
			wfq=new PacketWFQ[buckets_];
			return (TCL_OK);
		}
		if (strcmp(argv[1],"mask")==0) {
			mask_= atoi(argv[2]);
			return (TCL_OK);
		}
		if (strcmp(argv[1],"bandwidth")==0) {
			bandwidth_= atoi(argv[2]);
			return (TCL_OK);
		}
		
	}
	if(argc==4) {
		if (strcmp(argv[1],"quantum") == 0) {
			if(!wfq) {
				fprintf(stderr,"Set bucket size before trying to set quantum");
				exit(1);
			}
			if(atoi(argv[2]) >= buckets_) {
				fprintf(stderr,"quantum (= %d) cannot be greater then bucket size.\n", atoi(argv[2]));
				exit(1);
			}
			wfq[atoi(argv[2])].weight = atof(argv[3]);
			return (TCL_OK);
		}
	}
	return (Queue::command(argc, argv));
}

int WFQ::hash(Packet* pkt)
{
	hdr_ip *iph=hdr_ip::access(pkt);
	int i;
	if (mask_)
		i = (int)iph->saddr() & (0xfff0);
	else
		i = (int)iph->saddr();
	return ((i + (i >> 8) + ~(i>>4)) % ((2<<23)-1))+1; // modulo a large prime
}









